/*
    
    Device_Comm.cpp - Suunto Dive Computer Base Communication Class Routines
    
    Copyright (C) 2008  by James Connell

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <trace.h>
#include "Headers\sAPI_Error.h"
#include "Headers\sAPI_Comm.h"
#include "Headers\sAPI_Helper.h"

namespace sAPI  {

const struct COMMOPTIONS Device_Comm::CommOption1 = {CBR_2400,DTR_CONTROL_ENABLE,false,ODDPARITY,1800 /*1500*/,40,40,false,mmMEMORYMODEL_1};
const struct COMMOPTIONS Device_Comm::CommOption2 = {CBR_9600,DTR_CONTROL_DISABLE,true,NOPARITY,1100,0x100,0x100,true,mmMEMORYMODEL_2};
const struct COMMOPTIONS Device_Comm::CommOption3 = {CBR_9600,DTR_CONTROL_ENABLE,false,NOPARITY,4400,0x100,0x100,false,mmMEMORYMODEL_3};  // from 3400

const int Device_Comm::Command_Read = 0x05;
const int Device_Comm::Command_Write = 0x06;
const int Device_Comm::Command_PreWrite = 0x07;
const int Device_Comm::Command_FirstDive = 0x08;
const int Device_Comm::Command_NextDive = 0x09;
const int Device_Comm::Command_Command = 0xA5;
const int Device_Comm::Command_Unknown = 0x0A;  // believe this is the Gekko reset command
const int Device_Comm::Command_Reset = 0x0B;    // D9/D6 (also cobra2/3 and vyper 2/3) reset command

const wchar_t * const ComStateError = L"Error Setting Comm Port";


Device_Comm::Device_Comm(const wchar_t *port,struct COMMOPTIONS CO, volatile const bool& t, size_t tries)
				: fPortOpen(false),fTerminate(t), fCommDelay(6),
				fCommPort(INVALID_HANDLE_VALUE),fRead(CO.Read),
				fMemoryModel(CO.Memory),fEventMask(EV_TXEMPTY),
				fState(_READ), fTry(tries)
 {
		   // set up comport
	CommandBuffer = new unsigned char [CO.CommandBufferSize];
	DataBuffer = new unsigned char [CO.DataBufferSize];

  #if defined(USE_FILE_FOR_DEBUG)   // ADDED for 'file' reader support ONLY
	fPortOpen = true;
  #else

	if(port == 0)
	 {
		Trace(trace::TraceError,"Device_Comm::Create() Unable To Open Port <No Port Specified>");
		throw sAPI::CriticalError(NO_PORT_SPECIFED);
	 }
	   fCommPort = ::CreateFileW(port,
					GENERIC_READ | GENERIC_WRITE,
					0,    // must be opened with exclusive-access
					NULL, // no security attributes
					OPEN_EXISTING, // must use OPEN_EXISTING
					0,    // not overlapped I/O
					NULL  // hTemplate must be NULL for comm devices
					);
		 if( 0 == fCommPort.valid() )
		  {
			  Trace(trace::TraceError,"Device_Comm::Create() Unable To Open Port <Port Busy or Not Present>");
			  throw sAPI::CommError(BAD_OPEN,L"Comm Port Not Found");
		  }

			// get Old DCB and Old Timeouts
		 OldCommControlBlock.DCBlength = sizeof (DCB);
		 if( (0 == GetCommState(fCommPort, &OldCommControlBlock))
			 || (0 == GetCommTimeouts( fCommPort, &OldCommTimeOutControl )) )
		  {
			  //CloseHandle(fCommPort);
			  fCommPort.close();
			  Trace(trace::TraceError,"Device_Comm::Create() Can't Get Comm state or timeouts");
			  throw sAPI::CommError(BAD_OPEN,L"Invalid Port Specified");
		  }

		 Trace(trace::TraceDetail,"Device_Comm::Create() Port Opened (%d)",fCommPort);

		 CommControlBlock.DCBlength = sizeof (DCB);
		 if( 0 == ::GetCommState(fCommPort,&CommControlBlock))
		  {
			  //CloseHandle(fCommPort);
			  fCommPort.close();
			  Trace(trace::TraceError,"Device_Comm::Create() Can't Fill DCB with old Comm State");
			  throw sAPI::CriticalError(ComStateError);
		  }

		 //CommControlBlock.BaudRate;
		 CommControlBlock.fBinary = true;               // Binary mode; no EOF check
		 CommControlBlock.fParity = false;
		 CommControlBlock.fOutxCtsFlow = false;
		 CommControlBlock.fOutxDsrFlow = false;
		 //CommControlBlock.fDtrControl;
		 CommControlBlock.fDsrSensitivity = false;
		 CommControlBlock.fTXContinueOnXoff = false;      // XOFF continues Tx
		 CommControlBlock.fOutX = false;                // No XON/XOFF out flow control
		 CommControlBlock.fInX = false;                 // No XON/XOFF in flow control
		 CommControlBlock.fErrorChar = false;           // Disable error replacement
		 CommControlBlock.fNull = false;                // Disable null stripping
		 CommControlBlock.fRtsControl = RTS_CONTROL_DISABLE;
		 //CommControlBlock.fAbortOnError;  // Do not abort reads/writes on
		 CommControlBlock.XonLim = 2048;
		 CommControlBlock.XoffLim = 512;
		 CommControlBlock.ByteSize = 8;
		 //CommControlBlock.Parity ;
		 CommControlBlock.StopBits = ONESTOPBIT;
		 CommControlBlock.XonChar = 0x11;
		 CommControlBlock.XoffChar = 0x13;
		 CommControlBlock.ErrorChar = 0x00;
		 CommControlBlock.EofChar = 0x00;
		 CommControlBlock.EvtChar = 0x00;


		 CommControlBlock.BaudRate = CO.BaudRate;
		 CommControlBlock.fDtrControl = CO.fDtrControl;
		 CommControlBlock.fAbortOnError = CO.fAbortOnError;
		 CommControlBlock.Parity = CO.Parity;

		 if( 0 == ::SetCommState(fCommPort,&CommControlBlock))
		  {
			  //CloseHandle(fCommPort);
			  fCommPort.close();
			  Trace(trace::TraceError,"Device_Comm::Create() Can't Set Comm state");
			  throw sAPI::CriticalError(ComStateError);
		  }
		 if( 0 == ::GetCommTimeouts( fCommPort, &CommTimeOutControl ))
		  {
			  //CloseHandle(fCommPort);
			  fCommPort.close();
			  Trace(trace::TraceError,"Device_Comm::Create() Can't Get Timeouts");
			  throw sAPI::CriticalError(ComStateError);
		  }
		 CommTimeOutControl.ReadIntervalTimeout =  0;
		 CommTimeOutControl.ReadTotalTimeoutMultiplier = 0;
		 CommTimeOutControl.ReadTotalTimeoutConstant = CO.ReadTotalTimeoutConstant;
		 CommTimeOutControl.WriteTotalTimeoutMultiplier = 0;
		 CommTimeOutControl.WriteTotalTimeoutConstant = 1000;

		 if( 0 == ::SetCommTimeouts( fCommPort, &CommTimeOutControl ))
		  {
			  ::SetCommState(fCommPort,&OldCommControlBlock);
			  Trace(trace::TraceError,"Device_Comm::Create() Can't Set New Timeouts");
			  //CloseHandle(fCommPort);
			  fCommPort.close();
			  throw sAPI::CriticalError(ComStateError);
		  }

		 ::GetCommMask(fCommPort,&fOldEventMask);
		 fEventMask = EV_TXEMPTY;
		 if( 0 == ::SetCommMask(fCommPort, EV_TXEMPTY))
		  {
			  ::SetCommState(fCommPort,&OldCommControlBlock);
			  Trace(trace::TraceError,"Device_Comm::Create() Can't Set CommMask");
			  //CloseHandle(fCommPort);
			  fCommPort.close();
			  throw sAPI::CriticalError(ComStateError);
		  }

		 if( 0 == ::SetupComm(fCommPort,32640,1280))
          {
              ::SetCommState(fCommPort,&OldCommControlBlock);
			  Trace(trace::TraceError,"Device_Comm::Create() Invalid Setup");
			  //CloseHandle(fCommPort);
 			  fCommPort.close();
			  throw sAPI::CriticalError(ComStateError);
		  }
		 fPortOpen =true;
		 SetDTR(true);
  #endif
	return;
 }

Device_Comm::~Device_Comm()
 {
 	Trace(trace::TraceDetail,"Device_Comm::Destroy()");
	delete [] CommandBuffer;
 	delete [] DataBuffer;
	Close();
 }

void Device_Comm::Close(void) throw()
 {
	if(fPortOpen==false) return;
	::SetCommMask(fCommPort,fOldEventMask);
    ::SetCommState(fCommPort,&OldCommControlBlock);
	::SetCommTimeouts( fCommPort, &OldCommTimeOutControl );
	fCommPort.close();
	fPortOpen = false;
 }

int Device_Comm::delay(int val) throw()
 {
    if(val == 0 ) return fCommDelay;
    if(val < 1 ) val = 1;
    else if(val > 12) val = 12;
    return (fCommDelay = val);
 }


void Device_Comm::SetRTS( bool Val ) throw(sAPI::UnknownError)
 {
	DWORD z(CLRRTS);
	if(Val) z = SETRTS;
 	//else z = CLRRTS;
	if(0==::EscapeCommFunction(fCommPort,z)) throw sAPI::UnknownError(L"SetRTS");
 }

void Device_Comm::SetDTR(bool Val ) throw(sAPI::UnknownError)
 {
	DWORD z(CLRDTR);
	if(Val) z = SETDTR;
	//else z = CLRDTR;
	if(0==::EscapeCommFunction(fCommPort,z))  throw sAPI::UnknownError(L"SetDTR");
 }

unsigned char Device_Comm::getchecksum( const unsigned char *package, int len ) throw(sAPI::CriticalError)
 {
	unsigned char checksum(0);
	if(len < 1) throw sAPI::CriticalError(L"Attempt to checksum Null buffer");
	for( ; len > 0; --len) checksum ^= *package++;
	return checksum;
 }

int Device_Comm::checkpackagechecksum(unsigned char *dataptr,int Count) throw()
 {
	unsigned char c;
	try{
		c = getchecksum(dataptr,Count); // builds a check sum for command + Count + data[count]
	 }
	catch(...) { return -1; } // change this if needed to rethrow
	return ( c - dataptr[Count]);  // return ZERO if Equal - nonzero if bad CRC or Error
 }
}

