
/*
    
    sAPI_Comm.h -  Suunto Dive Computer Communication Classes
    
    Copyright (C) 2008  by James Connell

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/


#ifndef sAPI_CommH
#define sAPI_CommH

#ifdef _DEBUG
     #include <iostream>
     #include <fstream>
#endif

#include <boost/utility.hpp>


#include "Headers/sAPI_Error.h"
#include "Headers/sAPI_Helper.h"
#include "SmartHandle.h"

#define _WRITE 1
#define _READ 0

namespace sAPI  {


struct COMMOPTIONS
 {
    DWORD BaudRate;
    DWORD fDtrControl;
    DWORD fAbortOnError;
    BYTE  Parity;
	DWORD ReadTotalTimeoutConstant;
	int CommandBufferSize;
	int DataBufferSize;
	bool Read;
    enum MemoryModel Memory;
 };

class Device_Comm : boost::noncopyable
 {
	private:
		volatile const bool& fTerminate;
		size_t fTry;
		CFileHandle fCommPort;
		//CHandle<HANDLE, CCloseHandle, INVALID_HANDLE_VALUE> fCommPort;
		bool fPortOpen;
		int fCommDelay;
		bool fRead;
		bool fState;
		DWORD fOldEventMask;
		DCB CommControlBlock;
		DCB OldCommControlBlock;
		COMMTIMEOUTS CommTimeOutControl;
		COMMTIMEOUTS OldCommTimeOutControl;
		enum MemoryModel fMemoryModel;

	protected:
		virtual void SetCommState(bool state) throw() =0;
		virtual void Talk(int b) throw(sAPI::Error) =0;
		virtual void Listen(int b) throw(sAPI::Error) =0;
		void SetRTS( bool Val ) throw (sAPI::UnknownError);
		void SetDTR(bool Val ) throw (sAPI::UnknownError);
		COMSTAT  ComStat;
		DWORD ComErrorMask; //CE_BREAK | CE_FRAME | CE_IOE | CE_RXPARITY | CE_OVERRUN | CE_RXOVER | CE_TXFULL;
		DWORD fEventMask;
		unsigned char getchecksum( const unsigned char *package, int len ) throw(sAPI::CriticalError);
		int checkpackagechecksum(unsigned char *dataptr,int Count) throw();
		size_t Tries(void) const throw() { return fTry; }
		const bool terminate(void) const throw() { return fTerminate; }

	public:
		Device_Comm(const wchar_t *port,struct COMMOPTIONS CO, volatile const bool& t=false, size_t tries =3);
		virtual ~Device_Comm();

		unsigned char *CommandBuffer;
		unsigned char *DataBuffer;
		inline HANDLE commport() const throw() { return fCommPort(); }
		inline bool commopen(void)const throw() { return fPortOpen; }
		int delay(int d) throw();
		void setstate(bool s) throw() { fState = s; }
		bool getstate(void) const throw() { return fState; }
		void inline Purge(void) const throw() { ::PurgeComm(fCommPort(),PURGE_RXCLEAR | PURGE_TXCLEAR); }
		virtual void Close(void) throw();
        virtual void Communicate(int s, int r) throw(sAPI::Error) =0;
		const static int Command_Read;
		const static int Command_PreWrite;
		const static int Command_Write;
		const static int Command_FirstDive;
		const static int Command_NextDive;
		const static int Command_Command;
		const static int Command_Unknown;
		const static int Command_Reset;
		const static struct COMMOPTIONS CommOption1;
		const static struct COMMOPTIONS CommOption2;
		const static struct COMMOPTIONS CommOption3;
 };

class CommModel1 : public sAPI::Device_Comm
 {
    private:
        bool fState;
		void SetCommState(bool state) throw();

	protected:
		int GetDataPackage(unsigned char *Buffer,int flag) throw(sAPI::CommError);;
		virtual void Talk(int b) throw(sAPI::Error);
		virtual void Listen(int b) throw(sAPI::Error);

	public:
		CommModel1(const wchar_t *port, volatile const bool& t=false, int d=6);
		virtual ~CommModel1();
		void Communicate(int s, int r) throw(sAPI::Error) { Talk(s); Listen(r); }
 };

class CommModel2 : public sAPI::Device_Comm
 {
	private:
		size_t fTry;
		bool fState;
		void SetCommState(bool state) throw();

	protected:
		void GetDeviceHeader(unsigned char *data);
		virtual void Talk(int b) throw(sAPI::Error);
		virtual void Listen(int b) throw(sAPI::Error);

	public:
		CommModel2(const wchar_t *port, volatile const bool& t=false, size_t tries =3);
		virtual ~CommModel2();
		void Communicate(int s, int r) throw(sAPI::Error);
 };

class CommModel3 : public sAPI::Device_Comm
 {
	private:
		size_t fTry;
		bool fState;
		void SetCommState(bool state) throw();

	protected:
		void GetDeviceHeader(unsigned char *data);
		virtual void Talk(int b) throw(sAPI::Error);
		virtual void Listen(int b) throw(sAPI::Error);

	public:
		CommModel3(const wchar_t *port, volatile const bool& t=false, size_t tries =3);
		~CommModel3();
		void Communicate(int s, int r) throw(sAPI::Error);
 };

#ifdef _DEBUG

class CommModel2F : public sAPI::Device_Comm
 {
    private:
        bool fState;
		void SetCommState(bool state) throw();
		virtual void Talk(int b) throw(sAPI::Error);
		virtual void Listen(int b) throw(sAPI::Error);
		std::ifstream infile;
		bool fCommOpen;
		unsigned char *file;
		unsigned char AddressHigh;
		unsigned char AddressLow;
		unsigned char AddressSize;
		const wchar_t * fPort;

    protected:

    public:
		CommModel2F(const wchar_t *port, volatile const bool& t=false, size_t tries =3);
		~CommModel2F();
		void __stdcall Open(int z) throw(sAPI::Error);
        void Communicate(int s, int r) throw(sAPI::Error) { Talk(s); Listen(r); }
 };
 
class CommModel3F : public sAPI::Device_Comm
 {
    private:
        bool fState;
		void SetCommState(bool state) throw();
		virtual void Talk(int b) throw(sAPI::Error);
		virtual void Listen(int b) throw(sAPI::Error);
		std::ifstream infile;
		bool fCommOpen;
		unsigned char *file;
		unsigned char AddressHigh;
		unsigned char AddressLow;
		unsigned char AddressSize;
		const wchar_t * fPort;

    protected:

    public:
		CommModel3F(const wchar_t *port, volatile const bool& t=false, size_t tries =3);
		~CommModel3F();
		void __stdcall Open(int z) throw(sAPI::Error);
        void Communicate(int s, int r) throw(sAPI::Error) { Talk(s); Listen(r); }
 };
#endif


} // end namespace

#endif
