
/*
    
    Error.h -  The Exception Classes
    
    Copyright (C) 2008  by James Connell

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ErrorH
#define ErrorH

#include <wchar.h>
#include <stdlib.h>
#include <stdio.h>
#include <windows.h>
#include <wstdexcept.h>

namespace sAPI  {

enum ErrorCode { NoError=0, UNKNOWN_ERROR , CRITICALERROR, RUNTIME_ERROR, NOMEMORY, BUFFER_OVERRUN, BAD_WRITE_PARAM,
BAD_READ_PARAM, BAD_CHECKSUM, DATA_MISMATCH, INVALID_MODEL, INVALID_UNITS, BAD_OPEN, COMMAND_MISMATCH,
ERROR_SEND, ERROR_RECEIVE, BADSAMPLERATE, DEVICE_NOT_PRESENT, TIMEOUT, STACKUNDER, STACKOVER,
COMM_NOT_OPEN, NO_PORT_SPECIFED, ADDRESS_OUT_OF_RANGE, BADINSTRUCTION, ACTIONCANCELED = 9999 };


class Error : public w_std::runtime_error
 {
	//private:
	protected:
		ErrorCode _C;
		static const wchar_t* const message[];

	public:
		Error(const wchar_t * const _Message)
			 : w_std::runtime_error(_Message), _C(RUNTIME_ERROR)
		 {}
		Error(const wchar_t * const _Message, ErrorCode _Code )
			 : w_std::runtime_error(_Message), _C(_Code)
		 {}
		Error( ErrorCode _Code )
			 : w_std::runtime_error(message[_Code]), _C(_Code)
		 {}

		Error(const Error& _Right) throw()
		  : w_std::runtime_error(_Right._Ptr), _C(_Right._C)
		 {}

		Error& operator=(const Error& _Right) throw()
		 {  _Ptr = _Right._Ptr; _C = _Right._C; return (*this);  }

		virtual ~Error() throw() {}

		ErrorCode code(void) const { return _C; }
 };

class Device_Not_Present : public Error
 {
	public:
		Device_Not_Present(const wchar_t * const msg)
		   : Error(msg,DEVICE_NOT_PRESENT) {}
		~Device_Not_Present() throw() {}
 };

class User_Cancel : public Error
 {
	private:
	    static const wchar_t * const _CS;
	public:
		User_Cancel()
		   : Error(_CS,ACTIONCANCELED) {}
		~User_Cancel() throw() {}
 };

class CommError : public Error
 {
	public:
		CommError(ErrorCode _Code,const wchar_t * const msg)
		   : Error(msg,_Code) {}
		CommError(ErrorCode _Code)
		   : Error(_Code) {}
		~CommError() throw() {}
 };

class UnknownError : public Error
 {
	private:
		wchar_t message[100];
		size_t len;
		const wchar_t * const _SE(const wchar_t * const _A) throw()
		 {
			len = wcslen(_A);
			if(len >70)
			 {
				message[70]= '\0';
				len = 70;
			 }
			wcscpy(message,_A);
			swprintf(&message[len],29,L" : %d",::GetLastError());
			return this->message;
		 }
		 
	protected:
		const wchar_t* const getmsg(void) const throw() { return this->message; }
		void setmsg(const wchar_t * const msg) throw() { wcsncpy(this->message,msg,99); message[99]='\0'; }

	public:
		UnknownError(const wchar_t * const msg)
		 : Error(_SE(msg),UNKNOWN_ERROR) {}

		UnknownError(CONST UnknownError& _Right)
		  : Error(_Right.message,UNKNOWN_ERROR) { this->setmsg(_Right.message); }

		UnknownError& operator=(const UnknownError& _Right) throw()
		 {
			this->setmsg(_Right.message);
			_Ptr = _Right._Ptr;
			_C = _Right._C;
			return (*this);
		 }

		~UnknownError() throw() {}
 };

class StackError : public Error
 {
	public:
		StackError(ErrorCode _C)
		   : Error(_C) {}
		~StackError() throw() {}
 };

class StackEmpty : public StackError
 {
	public:
		StackEmpty()
		   : StackError(STACKUNDER) {}
		~StackEmpty() throw() {}
 };

class CommPartialTimeout : public Error
 {
	public:
		CommPartialTimeout()
			: Error(TIMEOUT) {}
		~CommPartialTimeout() throw() {}
 };

class BadDiveFormat : public Error
 {
	public:
		BadDiveFormat()
			: Error(L"Bad Format Returned") {}
		~BadDiveFormat() throw() {}
 };

class CriticalError : public Error
 {
	public:
		CriticalError(const wchar_t * const msg)
		   : Error(msg,CRITICALERROR) {}
		CriticalError(ErrorCode _Code)
		   : Error(message[_Code],CRITICALERROR) {}
		~CriticalError() throw() {}
 };

class MemoryRangeError : public Error
 {
	public:
		MemoryRangeError() : Error(ADDRESS_OUT_OF_RANGE)  {}
		~MemoryRangeError() throw() {}
 };
 
class BadInstruction : public Error
 {
	public:
		BadInstruction(enum ErrorCode ec = BADINSTRUCTION) : Error(ec)  {}
		~BadInstruction() throw() {}
 };
}
#endif
