/*
    
    DownloadControler.cpp -  Type 2+ download classes
    
    Copyright (C) 2008  by James Connell

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <mem.h>
#include <fstream>
#include <trace.h>

#include "headers\DownloadControler.h"
#include "headers\sAPI_Error.h"


const unsigned MemoryBase(MEMORYBASE);
const unsigned MemoryBaseOffset(BASEOFFSET);
const unsigned MemoryStart(0x19A);
const size_t HeaderSize(HEADERSIZE);
const size_t PageSize(PAGESIZE);
const int MaxPage(269);



TDiveDownloader::TDiveDownloader(sAPI::Device *D, PROFILEPTRS & P)
		: fDevice(D), RingBuffer(new TRingBuffer(D)),
		  ptrcurrent(P.CurrentDive), ptrempty(P.EmptyDive),
		  ptroldest(P.OldestDive),ptrnewest(P.NewestDive), fDivecount(P.DiveCount)
 {
	fSize = RingBuffer->size(ptroldest,ptrempty);
	Trace(trace::TraceDetail,"DownloadControl::Create.RingBuffer() , size = 0x%X",fSize);
	fDevice->Progress = fSize;
 }

TDiveDownloader::~TDiveDownloader()
 {
	delete RingBuffer;
 }


bool TDiveDownloader::GetDive(DIVE D) throw(sAPI::Error)
 {
	Dive = D;
	if(!fDivecount)
	 {
		Trace(trace::TraceInfo,"DownLoadControl:: No Dives");
		return false; // no dive to parce
	 }
	if(ptrcurrent == ptrempty)
	 {
		Trace(trace::TraceInfo,"DownLoadControl:: Finished");
	    return false; // finished
	 }
	if(fDevice->TerminateRaised()) throw sAPI::User_Cancel();
	fetchheader(ptrcurrent);
	fDevice->Progress += HeaderSize;
	if(fDevice->TerminateRaised()) throw sAPI::User_Cancel();
	setheaderdata();
	fetchprofile();
	ptrcurrent = ptrnewer;
	if(fDevice->TerminateRaised()) throw sAPI::User_Cancel();
	return true; //dive successfuly parced
 }

void TDiveDownloader::fetchheader(unsigned address) throw()// paramater Needed????
 {
	// fetch first size then determine final size

	Trace(trace::TraceInfo,"DownLoadControl::FetchHeader() ");
	RingBuffer->read(ptrcurrent,HeaderSize);
	fHeader = RingBuffer->data();
	scanfordata();
	ptrnewer= MakeWord(&fHeader[2]);
	ptrolder= MakeWord(&fHeader[0]);
	Trace(trace::TraceDetail,"DownLoadControl:: Newer @ &0x%X Older @ &0x%X",ptrnewer,ptrolder);
 }

void TDiveDownloader::scanfordata(void) throw(sAPI::Error)
 {
	fHeaderSize = 0x4E;
	for(int i(0x43); i<0x4E; i++)
	 {
		 if( (*(fHeader+i)== 0x80) && (*(fHeader+i+1)== 0x01) )
		  {
			  fHeaderSize = i+6;
			  Trace(trace::TraceDetail,"DownLoadControl:: Header Size = %u",fHeaderSize);
			  return;
		  }
	 }
	Trace(trace::TraceError,"DownLoadControl:: Header Size Error - End Not Found ");
	throw(sAPI::CriticalError(L"DownloadControl::Header Size Unknown"));
 }

void TDiveDownloader::fetchprofile(void) throw()
 {
	Trace(trace::TraceInfo,"DownLoadControl::FetchProfile(): begin");
	size_t profilelen(RingBuffer->size(ptrcurrent+fHeaderSize,ptrnewer));
	Trace(trace::TraceDetail,"DownLoadControl::FetchProfile(): ProfileLength = %u",profilelen);
	RingBuffer->read(ptrcurrent+fHeaderSize,profilelen);
	Dive->Profile = new unsigned char[profilelen];
	Dive->ProfileLength = profilelen;
	memcpy(Dive->Profile,RingBuffer->data(),profilelen);
 }

void TDiveDownloader::setheaderdata(void) throw()
 {
    Dive->PressureIncluded = (fHeaderSize < 0x4D ? false : true);
	Dive->MaxDepth = MakeWord(&fHeader[0xD]);
	Dive->DiveTime = MakeWord(&fHeader[0x0F]);
	Dive->CNSClock = fHeader[0x13];
	Dive->PPO2 = 0x7F;
	Dive->SampleInterval = fHeader[0x1C];
	Dive->DiveNumber = fHeader[0x1E];
	Dive->SurfaceIntervalMinutes   = fHeader[0x20 ];
	Dive->SurfaceIntervalHours = fHeader[0x1F];
	Dive->ConservationSettings = (fHeader[0x22]*3) + fHeader[0x23];
	Dive->Year = (MakeWord(&fHeader[0x18])%2000);
	Dive->Month = fHeader[0x1A];
	Dive->Day = fHeader[0x1B];
	Dive->Hour = fHeader[0x15];
	Dive->Minute = fHeader[0x16];
	Dive->Second = fHeader[0x17];
	Dive->GasModel = fHeader[0x1D];
	Dive->RGBM = fHeader[0x21];
	Dive->GasMix = fHeader[0x25];
	Dive->GasMix2 = fHeader[0x26];
	Dive->GasMix3 = fHeader[0x27];
	Dive->TempInterval = fHeader[fHeaderSize-7];
	Dive->EventMarker = MakeWord(&fHeader[fHeaderSize-2]);
	// build tissue vector
	for(int i(0); i < 9; ++i)
     {
		Dive->tissue[i] = (MakeWord(&fHeader[0x29+(i*2)]));
	 }
 }

 
/* ----------  TRingBuffer  ------------------ */

size_t TRingBuffer::size(size_t start, size_t end) const throw()
 {
	size_t len(end - start); 
    if( // ToDo:  newest error seems to be here! in D9 7/5/09
		( (start > fDevice->MemoryTop()) || (end > fDevice->MemoryTop()) )
	   ||
		( (start < MemoryBaseOffset) || (end < MemoryBaseOffset) )
	  )
     {
        Trace(trace::TraceError,"RingBuffer::size() Len=0 start=%u end=%u",start,end);
     #ifdef USE_TRACE   
		if( (start > fDevice->MemoryTop()) )
		   Trace(trace::TraceError,"start > MemoryTop");
        if( end > fDevice->MemoryTop() )
           	Trace(trace::TraceError,"End > MemoryTop");
		if(start < MemoryBaseOffset) Trace(trace::TraceError,"Start < MemoryBaseOffset");
		if(end < MemoryBaseOffset ) Trace(trace::TraceError,"End < MemoryBaseOffset");
     #endif
        return 0;
        }

	if( end < start) // wraps memory!
	 {
		len = fDevice->MemoryTop() - start;
		len += end - MemoryStart;
	 }
	return len;
 }

void TRingBuffer::read(unsigned address,int length) throw(sAPI::Error)
 {
	Trace(trace::TraceDetail,"RingBuffer::read() begin");
	if(length ==0)
	 {
		Trace(trace::TraceError,"RingBuffer::read() Length = 0");
		throw(sAPI::Error(L"RingBuffer: Zero Length Read"));
	 }

	delete[] fdata; // create new profile buffer
	fdata = new unsigned char [length];
	fptr = fdata;

	flength = 0; // no char read yet
	int remainder(length);  // set to read complete amount
	page->set(address);
	while(remainder > 0)
	 {
		page->fetch();
		if(remainder > page->bytesavailable() )
		 {
			int B(page->bytesavailable());
			fptr +=	page->get(fptr,B);
			remainder -= B;
			flength += B;
			fDevice->Progress += B;
			page->next();
		 }
		else
		 {
			fptr +=	page->get(fptr,remainder);
			flength += remainder;
			fDevice->Progress += remainder;
			remainder =0;
		 }
	 }
	Trace(trace::TraceDetail,"RingBuffer::read() Completed   amount = %u",flength);
 }

/* ---------- END  TRingBuffer  ------------------ */



int Page::bytesavailable(void) const throw()
 {
	return fBytes;
 }

void Page::set(unsigned Address) throw()
 {
	fAddress = Address;
	fPage = page(Address);
	fOffset = getoffset(Address);
	fBytes = PAGESIZE - fOffset;
 }

void Page::fetch(void) throw(sAPI::Error)
 {
    if(fDevice->TerminateRaised())
     {
		Trace(trace::TraceDetail,"Page::fetch() User Cancel ");
        throw sAPI::User_Cancel();
	 }
	int size(PAGESIZE);
	int currentaddress(getpageaddress(fPage));
	if(currentaddress > 0x7f00)
	{
		size = size;
	}
	if((/*getpageaddress(fPage)*/currentaddress + size ) > (int)fDevice->MemoryTop())
	 {
		size = fDevice->MemoryTop() - currentaddress /*getpageaddress(fPage)*/;
		Trace(trace::TraceDetail,"Page::fetch() adjusting for Memory Top new page size = %d",size);
	 }
	fBytes = size - fOffset;

    try {
		Trace(trace::TraceDetail,"Page::fetch() Reading profile @ address 0x%X ",getpageaddress(fPage)+fOffset);
		fDevice->Read(currentaddress/*getpageaddress(fPage)*/,size,_Pb);
		return;
     }
//	catch(sAPI::Error &e)
//	 {
//		Trace(trace::TraceError,"Page::fetch() Read Error : % ",e.what());
//		throw;
//	 }
	catch(...)
     {
		Trace(trace::TraceError,"Page::fetch() Read Error ");
        throw;
     }
 }

int Page::get(unsigned char *ptr,int bytes) throw()
 {
//	if( (_Pb+fOffset + bytes) > fDevice->MemoryTop())  )
//	 {
//		bytes = fDevice->MemoryTop() - (_Pb+fOffset + bytes)
//	 }
	memcpy(ptr,_Pb+fOffset,bytes);
	return bytes;
 }

void Page::next(void) throw()
 {
	if(fPage < MaxPage) set(getpageaddress(++fPage));
	else set(PROFILEBASE);
  }

