
/*
    
    DeviceType1.h -  Higher level routines
    
    Copyright (C) 2008  by James Connell

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <Trace.h>
#include "headers\sAPI_Devices.h"
#include "headers\GeneraterT1.h"
#include "headers\GeneraterT2.h"

namespace sAPI  {

Device_Type1::Device_Type1(const wchar_t * const port,volatile const bool& terminate,int delay)
	 : sAPI::Device_Type1_Control(port,terminate,delay), fLastDive(0), fIsFirstDive(true), S(new sAPI::Stack(0x2000))
 { }

Device_Type1::~Device_Type1()
 {
	delete S;
 }

 void Device_Type1::Open(int tries)  throw(User_Cancel,Device_Not_Present)
 {
	Device_Type1_Control::Open(tries);
	InitDownload();
 }

void Device_Type1::InitDownload(void)  throw(Error)
{
	fIsFirstDive = true;
	Dives->clear();
}


bool Device_Type1::DownloadDive(void) throw(Error)
 {
	 int result(0);
	 ADive =  DIVE(new DIVE_DATA);
	 if(fIsFirstDive)
	 {
		Progress[0x2000 - 0x72];
		fIsFirstDive = false;
		try {
			result = GetFirstDive(ADive);
		 }
		catch(const BadDiveFormat& e)
		 {
			Trace(trace::TraceError,"Device_Type1::DownloadDive(): Bad Dive Format! Ignoring Dive");
			return true;
		 }
		catch(Error& e)
		 {
			throw;
		 }
		if( result == 0 ) return false;
		Dives->push_front(ADive);
		return true;
	 }

	try {
		result = GetNextDive(ADive);
	 }
	catch(const BadDiveFormat& e)
	 {
		Trace(trace::TraceError,"Device_Type1::DownloadDive(): Bad Dive Format! Ignoring Dive");
		return true;
	 }
	catch(const Error& e)
	 {
		throw;
	 }

	 if( result == 0 )	return false;

	Dives->push_front(ADive);
	return true;     
 }

void Device_Type1::DownloadDevice(void) throw(Error)
 {
	unsigned char buffer[32];

	Progress[75];

	Read(0x1E,14,buffer);

	Progress += 14;
	memcpy(&DeviceInfo.MaxDepth,buffer,14);

	if(TerminateRaised() == true) throw(User_Cancel());

	Read(0x2C,32,buffer);

	Progress += 32;
	memcpy(&DeviceInfo.PersonalInfo,buffer,32);
	if(TerminateRaised() == true) throw(User_Cancel());

	Read(0x53,29,buffer);

	Progress += 29;
	memcpy(&DeviceInfo.SampleInterval,buffer,9);
	memcpy(&DeviceInfo.MaxFreeDiveDepth,buffer+9,21);

    swapbytes(DeviceInfo);
    
	return;
 }

void Device_Type1::MakeOutput(DCDMLDocument& Doc,unsigned OutputOptions) throw(Error)
 {
	setunits(OutputOptions);

	delete Output;
	switch(OutputOptions & ooVersion)
	 {
        case 0: // No Type specified
		case ooVersion1: //OUTPUTTYPE1:
			Output = new OutputGeneraterT1D1(&DeviceInfo,Dives,Model());
			break;
		case ooVersion2: //OUTPUTTYPE2:
		   	Output = new OutputGeneraterT2D1(&DeviceInfo,Dives,Model(),Units());
		 	break;
		default:
        	Output=0;
        	Trace(trace::TraceError,"Device_Type1::MakeOutput(): Requested invalid output Format");
			throw(Error(L"Output: format not supported"));
			//break;
	 }

    Doc << DCDML_USE_UTF8;

	if(OutputOptions & ooUseFullTags) Doc << DCDML_USE_FULLTAGS;
    else Doc >> DCDML_USE_FULLTAGS;

	if(OutputOptions & ooUseFormat) Doc << DCDML_USE_FORMATTING;
    else Doc >> DCDML_USE_FORMATTING;

	Output->Generate(Doc);
	delete Output;
	Output =0;
 }
 


/* ------------------------  Protected Functions ---------------------------*/



/* ------------------    Private Functions  -----------------*/

void Device_Type1::swapbytes(DEVICE_DATA& data) throw()
 {
    dobyteswap(&data.MaxDepth);
    dobyteswap(&data.TotalDiveTime);
    dobyteswap(&data.TotalDives);
    dobyteswap(&data.MaxFreeDiveDepth);
    dobyteswap(&data.MaxFreeDiveTime);
    dobyteswap(&data.AlarmTime);
    dobyteswap(&data.AlarmDepth);
 }

void Device_Type1::dobyteswap( unsigned short *D) throw()
 {
    unsigned short big, little;
    little = (unsigned short)((*D & 0xff00) >>8);
    big = (unsigned short)((*D & 0x00ff) <<8);
    *D = (unsigned short) (little |  big );
 }

bool Device_Type1::parce_divedata(DIVE data) throw(BadDiveFormat)
 {
    unsigned char c;
	int i(0);
	try
	 {
		data->SurfaceIntervalMinutes = S->pop();
		data->SurfaceIntervalHours = S->pop();
		data->DiveNumber = S->pop();
		data->SampleInterval = S->pop();
		data->ConservationSettings = S->pop();
		data->StartingPressure = S->pop()/2;
		data->GasMix = S->pop();
		data->PPO2 = S->pop();
		data->StartingAirTemp = S->pop();
		data->Year = S->pop();
		data->Month = S->pop();
		data->Day = S->pop();
		data->Hour = S->pop();
		data->Minute = S->pop();

		data->Profile = new unsigned char [S->size()+2];

		do
		 {
			c = S->pop();
			if(c != 0x80) data->Profile[i++] = c;
			if(c == 0x86)
			 {
				Trace(trace::TraceError,"ParceDive: End of Profile reached - No EOD Marker Found");
				throw( BadDiveFormat());
			 }
		 } while(c != 0x80);

		data->TempAtDepth = S->pop();
		data->EndingTemp = S->pop();
		data->EndingPressure = S->pop()/2;
		data->CNSClock = S->pop();
	 }
	catch(const StackEmpty &e)
	 { throw( BadDiveFormat() ); }

	data->ProfileLength = i;
		// added 8/24/07
	if(data->GasMix ==0)
	 {
		data->GasMix = 21;
		data->GasMix2 = 0;
		data->GasMix3 = 0;
		data->GasModel = 0;
	 }
	else data->GasModel = 1;
	// end add
	return true;
}

void  Device_Type1::GetaDivesData(unsigned char cmd) throw(Error)
 {
    unsigned char data[40];
    int count;
    bool flag(false);
	S->clear();
	CommandBuffer[0] = cmd;
    CommandBuffer[1] = Command_Command;
    std::memset(data,0,40);
	try
     {
     	Talk(2);
     }
	catch(const Error& e)
	 {
		Trace(trace::TraceError,"GetDiveData():: Exception thrown (1) - %s ",TraceString(e.what()));
		throw;
	 }
	S->push(0x86);
	for(;;)
	 {
		if(TerminateRaised() == true) throw(User_Cancel());
		try
		 {
			count = GetDataPackage(data,flag);
		 }
		catch(const Error& e)
		 {
			Trace(trace::TraceError,"GetDataPackage():: Exception thrown (2) - %s",TraceString(e.what()));
			throw;
         }

		if(count == 0 )
		 {
			if(flag == false)
             {
             	Trace(trace::TraceInfo,"End of Dives Found");
                fLastDive = true;
             }
			return;
		 }
        flag = true;
		Progress += count;
        for(int i(2); i < (count+2); ++i)
         {
			S->push(data[i]);
         }
        if(count < 32) break;
     }
  }

bool Device_Type1::GetFirstDive(DIVE Data) throw(Error)
 {
	::Sleep(666);
	GetaDivesData(Command_FirstDive);
    if(fLastDive ==true ) return false;
	return parce_divedata(Data);
 }

bool Device_Type1::GetNextDive(DIVE Data) throw(Error)
 {
	 // ToDo -o  -c Why did I do This???: is this nessessary????????
	::Sleep(666); // +(fCommDelay * 33));    // changed 9/13/07
	GetaDivesData(Command_NextDive);
	if(fLastDive) return false;
	return parce_divedata(Data);
 }
}
