
/*
    
    DeviceType3_Control.cpp -  Dive Computer Communication Routines
    
    Copyright (C) 2008  by James Connell

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/


#include <trace.h>
#include "headers\sAPI_Devices_Control.h"

namespace sAPI
{

const int min_packagesize(0x28);
const int max_packagesize(0x78);

Device_TypeC2V2_Control::Device_TypeC2V2_Control(const wchar_t * const port,volatile const bool& terminate)
	: Device(mmMEMORYMODEL_3,0x7FF7,terminate), CommModel3(port,terminate)
 { }

void Device_TypeC2V2_Control::Open(int tries) throw(User_Cancel,Device_Not_Present)
 {
	Progress[tries];
	Progress.maximum();
	for(int i(0); i < tries; ++i)
	 {
		try {
			getheader();
		 }
		catch(Error& e)
		 {
		    --Progress;
			if(TerminateRaised() == true)
			 {
				Trace(trace::TraceInfo,"Device_Type3::Open() User Cancel Detected");
				throw(User_Cancel());
			 }
			Sleep(500);
			continue;
		 }
		Status(L"Comm Opened Successfully!");
		Progress.minimum();
		Action(1);
		return;
	 }
	Action(0);
	throw( Device_Not_Present(L"Device Type3 Not Found") );
 }

 //Done 2:change this to use device specific strings
 
int Device_TypeC2V2_Control::Read(unsigned address,int bytes,BYTE *data,bool P)  throw(Error)
 {
	if( ( bytes < 1) || ( data == 0))
	 {
		Trace(trace::TraceError,"Device_Type3::Read() Bad Read Command");
		throw BadInstruction(BAD_READ_PARAM);
	 }
	if( static_cast<size_t>(address + bytes) > (MemoryTop()+1) )
	 {
		Trace(trace::TraceError,"Device_Type3::Read() Memory Overrun Last Address = 0x%X",static_cast<size_t>(address + bytes) );
		throw MemoryRangeError();
	 }

	int totalread(0);
 
	if( bytes < min_packagesize ) // Small Read
	 {
		unsigned page(address);
		if((address + min_packagesize) > MemoryTop() )  // need to adjust read offset
		 {
			page = MemoryTop() - min_packagesize;
			totalread = read_(page,min_packagesize,0);
			memcpy(data, &DataBuffer[6+(address - page)],bytes);
		 }
		else
		 {
			totalread = read_(page,min_packagesize,0);
			memcpy(data,&DataBuffer[6],bytes);
		 }
		if(P) Progress += totalread;
	 }
	else if( bytes > max_packagesize ) // Large Read
	 {
	 	Trace(trace::TraceDetail,"Device_Type3::Read() Large Read");
		unsigned add(address);
		const int pages(bytes / max_packagesize);
		for( int i(0); i < pages; ++i)
		 {
			int temp(read_(add,max_packagesize,data));
			if(P) Progress += temp;
			add += temp;
			data += temp;
			totalread += temp;
			if(TerminateRaised() == true)
			 {
				Trace(trace::TraceInfo,"Device_Type3::Read() User Cancel Detected");
				throw(User_Cancel());
			 }
		 }
		if((bytes - totalread) > 0 )
		 {
			int temp(Read(add,(bytes - totalread),data));
			totalread += temp;
			if(P) Progress += temp;
		 }
	 }
	else
	 {
		totalread = read_( address, bytes, data);
		if(P) Progress += totalread;
	 }
	return totalread;
 }

unsigned char Device_TypeC2V2_Control::ReadByte(unsigned address,bool P) throw(Error)
 {
	if( static_cast<size_t>(address) > MemoryTop())
			throw MemoryRangeError();
	unsigned char data;
	Read(address,1,&data,P);
	return data;
 }

unsigned short Device_TypeC2V2_Control::ReadWord(unsigned address,bool P) throw(Error)
 {
	if( static_cast<size_t>(address) > (MemoryTop()-1))
			throw MemoryRangeError();
	unsigned char data[2];
	unsigned short result;
	Read(address,2,data,P);
	result = data[0] <<8;
	result += data[1];
	return result;
 }

int Device_TypeC2V2_Control::Write(unsigned address,int bytes,const unsigned char *data,bool P) throw(Error)
 {
	if( ( bytes < 1) || ( data == 0)) throw BadInstruction(BAD_WRITE_PARAM);
	if(static_cast<size_t>(address + bytes) > (MemoryTop() + 1) ) throw MemoryRangeError();

	int totalwrite(0);

	if(bytes < min_packagesize)  // Small write
	 {
		unsigned page(address);
		if(address + min_packagesize > MemoryTop() )  // need to adjust read offset
		 {
			page = MemoryTop() - min_packagesize;
			read_(page,min_packagesize,0);
			memcpy(&DataBuffer[6+(address - page)],data,bytes);
		 }
		else
		 {
			read_(page,min_packagesize,0);
			memcpy(&DataBuffer[6],data,bytes);
		 }
		totalwrite = write_( page, bytes, &DataBuffer[6]);
		if(P) Progress += totalwrite;
	 }

	else if(bytes > max_packagesize)   // Large write
	 {
		int pages(bytes / max_packagesize);
		unsigned add(address);
		for( int i(0); i < pages; ++i)
		 {
			int temp(write_(add,max_packagesize,data));
			if(P) Progress += temp;
			add += temp;
			data += temp;
			totalwrite += temp;
			if(TerminateRaised() == true)
			 {
				Trace(trace::TraceInfo,"Device_Type3::Write() User Cancel Detected");
				throw(User_Cancel());
			 }
		 }
		if((bytes - totalwrite) > 0 )
		 {
			int temp(Write(add,(bytes - totalwrite),data));
			if(P) Progress += temp;
			totalwrite += temp;
		 }
	 }
	else
	 {
		totalwrite = write_( address, bytes, data);
		if(P) Progress += totalwrite;
	 }
	return totalwrite;
 }

//void Device_TypeC2V2_Control::WriteByte(unsigned address,unsigned char Byte,bool P) throw(Error)
// {
//	Write(address,1,&Byte,P);
// }

void Device_TypeC2V2_Control::WriteWord(unsigned address,unsigned short word,bool P) throw(Error)
 {
	unsigned char data[2];
	data[0] = HighByte(word);
	data[1] = LowByte(word);
	Write(address,2,data,P);
 }

int Device_TypeC2V2_Control::Command(unsigned send,const unsigned char *cmdstr,unsigned rec,unsigned char* data) throw(Error)
 {
	memcpy(CommandBuffer,cmdstr,send);
	Communicate(send,rec);
    if(rec && data) memcpy(data,&DataBuffer[6],rec);
    return rec;
 }


/* ------------------    Private Functions  -----------------*/
void Device_TypeC2V2_Control::getheader(void) throw(Error)
 {
	unsigned char DBuffer[0x30];
	static const unsigned char buffer[] = { 0xF,0,0  };
	Trace(trace::TraceDetail,"Header Fetch");
	try
	 {
		Command(3,buffer,8,0);
	 }
	catch(...)
	 {
		DeviceInfo.ComputerModel = 0;
		DeviceInfo.FirmwareVersionMjr = 0;
		DeviceInfo.FirmwareVersionMinor = 0;
		DeviceInfo.FirmwareVersionSubMinor = 0;
		Trace(trace::TraceError,"Device_TypeC2V2::getheader(): Exception Caught");
		throw;
	 }
	DeviceInfo.ComputerModel = DataBuffer[3];
	DeviceInfo.FirmwareVersionMjr = DataBuffer[4];
	DeviceInfo.FirmwareVersionMinor = DataBuffer[5];
	DeviceInfo.FirmwareVersionSubMinor = DataBuffer[6];
	Trace(trace::TraceDetail,"Device_TypeC2V2::Device ID = %x",DeviceInfo.ComputerModel);

	Sleep(100);

	Read(0x190,min_packagesize,DBuffer);

	DeviceInfo.ProfilePtrs.DiveCount = MakeWord(&DBuffer[2]);
	DeviceInfo.ProfilePtrs.OldestDive = MakeWord(&DBuffer[6]);
	DeviceInfo.ProfilePtrs.NewestDive = MakeWord(&DBuffer[0]);
	DeviceInfo.ProfilePtrs.EmptyDive = MakeWord(&DBuffer[4]);
	DeviceInfo.ProfilePtrs.CurrentDive = MakeWord(&DBuffer[6]);
 }

#if defined(BORLAND)
	#pragma warn -8057
#endif

int Device_TypeC2V2_Control::read_(unsigned address,int bytes,BYTE *data) throw(Error)
 {
	Sleep(500);
	CommandBuffer[0] = (unsigned char)Command_Read;
	CommandBuffer[1] = 0;
	CommandBuffer[2] = 3;
	CommandBuffer[3] = HighByte(address);
	CommandBuffer[4] = LowByte(address);
	CommandBuffer[5] = (unsigned char)bytes;
	try {
		Communicate(6,bytes+7);
	 }
	catch(...)
	 {
		Trace(trace::TraceError,"Device_TypeC2V2::Read(): Exception Raised");
		Trace(trace::TraceDetail,"Device_TypeC2V2::Read(): %d bytes at address 0x%x",bytes,address);
		throw;
	 }
	if(data) std::memcpy(data,&DataBuffer[6],bytes);
	return bytes;
 }

int Device_TypeC2V2_Control::write_(unsigned address,int bytes,const BYTE *data) throw(Error)
 {
	CommandBuffer[0] = (unsigned char)Command_Write;
	CommandBuffer[1] = 0;
	CommandBuffer[2] = 3 + (unsigned char)bytes;
	CommandBuffer[3] = HighByte(address);
	CommandBuffer[4] = LowByte(address);
	CommandBuffer[5] = (unsigned char)bytes;
	std::memcpy(&CommandBuffer[6],data,bytes);
	try {
        Communicate(6+bytes,7);
	 }
	catch(...)
	 {
		Trace(trace::TraceError,"Device_TypeC2V2::Write(): Exception Raised");
		Trace(trace::TraceDetail,"Device_TypeC2V2::Write():  %u bytes at address %u",bytes,address);
		throw;
	 }
	return bytes;
 }

#if defined(BORLAND)
	#pragma warn +8057
#endif
}
